@@ -23,10 +23,13 @@
 #include "config.h"
 
 #include <errno.h>
+#include <limits.h>
 #include <pwd.h>
 #include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 
 #include <glib.h>
 #include <glib-object.h>
@@ -49,6 +52,10 @@ struct _GdmDisplayAccessFilePrivate
 #define GDM_DISPLAY_ACCESS_COOKIE_SIZE 16
 #endif
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 static void gdm_display_access_file_finalize (GObject * object);
 
 enum
@@ -202,7 +209,7 @@ _get_uid_and_gid_for_user (const char *username,
         g_assert (gid != NULL);
 
         errno = 0;
-        passwd_entry = getpwnam (username);
+        gdm_get_pwent_for_name (username, &passwd_entry);
 
         if (passwd_entry == NULL) {
                 return FALSE;
@@ -214,33 +221,76 @@ _get_uid_and_gid_for_user (const char *username,
         return TRUE;
 }
 
+static void
+clean_up_stale_auth_subdirs (void)
+{
+        GDir *dir;
+        const char *filename;
+
+        dir = g_dir_open (GDM_XAUTH_DIR, 0, NULL);
+
+        if (dir == NULL) {
+                return;
+        }
+
+        while ((filename = g_dir_read_name (dir)) != NULL) {
+                char *path;
+
+                path = g_build_filename (GDM_XAUTH_DIR, filename, NULL);
+
+                /* Will only succeed if the directory is empty
+                 */
+                g_rmdir (path);
+                g_free (path);
+        }
+        g_dir_close (dir);
+}
+
 static FILE *
 _create_xauth_file_for_user (const char  *username,
                              char       **filename,
                              GError     **error)
 {
         char   *template;
+        const char *dir_name;
+        char   *auth_filename;
         int     fd;
         FILE   *fp;
         uid_t   uid;
         gid_t   gid;
 
-        fp = NULL;
+        g_assert (filename != NULL);
 
-        template = g_strdup_printf (GDM_XAUTH_DIR
-                                    "/auth-cookie-XXXXXXXX-for-%s",
-                                    username);
+        *filename = NULL;
 
-        fd = g_mkstemp (template);
-        *filename = template;
         template = NULL;
+        auth_filename = NULL;
+        fp = NULL;
+        fd = -1;
 
-        if (fd < 0) {
-                g_set_error (error,
-                             G_FILE_ERROR,
-                             g_file_error_from_errno (errno),
-                             "%s", g_strerror (errno));
-                goto out;
+        /* Create directory if not exist, then set permission 0711 and ownership root:gdm */
+        if (g_file_test (GDM_XAUTH_DIR, G_FILE_TEST_IS_DIR) == FALSE) {
+                g_unlink (GDM_XAUTH_DIR);
+                if (g_mkdir (GDM_XAUTH_DIR, 0711) != 0) {
+                        g_set_error (error,
+                                     G_FILE_ERROR,
+                                     g_file_error_from_errno (errno),
+                                     "%s", g_strerror (errno));
+                        goto out;
+                }
+
+                g_chmod (GDM_XAUTH_DIR, 0711);
+                _get_uid_and_gid_for_user (GDM_USERNAME, &uid, &gid);
+                if (chown (GDM_XAUTH_DIR, 0, gid) != 0) {
+                        g_warning ("Unable to change owner of '%s'",
+                                   GDM_XAUTH_DIR);
+                }
+        } else {
+                /* if it does exist make sure it has correct mode 0711 */
+                g_chmod (GDM_XAUTH_DIR, 0711);
+
+                /* and clean up any stale auth subdirs */
+                clean_up_stale_auth_subdirs ();
         }
 
         if (!_get_uid_and_gid_for_user (username, &uid, &gid)) {
@@ -249,24 +299,80 @@ _create_xauth_file_for_user (const char  *username,
                              GDM_DISPLAY_ERROR_GETTING_USER_INFO,
                              _("could not find user \"%s\" on system"),
                              username);
-                close (fd);
-                fd = -1;
                 goto out;
 
         }
 
+        template = g_strdup_printf (GDM_XAUTH_DIR
+                                    "/auth-for-%s-XXXXXX",
+                                    username);
+
+        g_debug ("GdmDisplayAccessFile: creating xauth directory %s", template);
+        /* Initially create with mode 01700 then later chmod after we create database */
+        errno = 0;
+        dir_name = gdm_make_temp_dir (template);
+        if (dir_name == NULL) {
+                g_set_error (error,
+                             G_FILE_ERROR,
+                             g_file_error_from_errno (errno),
+                             "Unable to create temp dir from tempalte '%s': %s",
+                             template,
+                             g_strerror (errno));
+                goto out;
+        }
+
+        g_debug ("GdmDisplayAccessFile: chowning %s to %u:%u",
+                 dir_name, (guint)uid, (guint)gid);
+        errno = 0;
+        if (chown (dir_name, uid, gid) < 0) {
+                g_set_error (error,
+                             G_FILE_ERROR,
+                             g_file_error_from_errno (errno),
+                             "Unable to change permission of '%s': %s",
+                             dir_name,
+                             g_strerror (errno));
+                goto out;
+        }
+
+        auth_filename = g_build_filename (dir_name, "database", NULL);
+
+        g_debug ("GdmDisplayAccessFile: creating %s", auth_filename);
+        /* mode 00600 */
+        errno = 0;
+        fd = g_open (auth_filename,
+                     O_RDWR | O_CREAT | O_EXCL | O_BINARY,
+                     S_IRUSR | S_IWUSR);
+
+        if (fd < 0) {
+                g_set_error (error,
+                             G_FILE_ERROR,
+                             g_file_error_from_errno (errno),
+                             "Unable to open '%s': %s",
+                             auth_filename,
+                             g_strerror (errno));
+                goto out;
+        }
+
+        g_debug ("GdmDisplayAccessFile: chowning %s to %u:%u", auth_filename, (guint)uid, (guint)gid);
+        errno = 0;
         if (fchown (fd, uid, gid) < 0) {
                 g_set_error (error,
                              G_FILE_ERROR,
                              g_file_error_from_errno (errno),
-                             "%s", g_strerror (errno));
+                             "Unable to change owner for '%s': %s",
+                             auth_filename,
+                             g_strerror (errno));
                 close (fd);
                 fd = -1;
                 goto out;
         }
 
-        fp = fdopen (fd, "w");
+        /* now open up permissions on per-session directory */
+        g_debug ("GdmDisplayAccessFile: chmoding %s to 0711", dir_name);
+        g_chmod (dir_name, 0711);
 
+        errno = 0;
+        fp = fdopen (fd, "w");
         if (fp == NULL) {
                 g_set_error (error,
                              G_FILE_ERROR,
@@ -276,7 +382,19 @@ _create_xauth_file_for_user (const char  *username,
                 fd = -1;
                 goto out;
         }
+
+        *filename = auth_filename;
+        auth_filename = NULL;
+
+        /* don't close it */
+        fd = -1;
 out:
+        g_free (template);
+        g_free (auth_filename);
+        if (fd != -1) {
+                close (fd);
+        }
+
         return fp;
 }
 
@@ -320,8 +438,13 @@ _get_auth_info_for_display (GdmDisplayAccessFile *file,
         gdm_display_is_local (display, &is_local, NULL);
 
         if (is_local) {
+                char localhost[HOST_NAME_MAX + 1] = "";
                 *family = FamilyLocal;
-                *address = g_strdup (g_get_host_name ());
+                if (gethostname (localhost, HOST_NAME_MAX) == 0) {
+                        *address = g_strdup (localhost);
+                } else {
+                        *address = g_strdup ("localhost");
+                }
         } else {
                 *family = FamilyWild;
                 gdm_display_get_remote_hostname (display, address, NULL);
@@ -485,17 +608,37 @@ gdm_display_access_file_remove_display (GdmDisplayAccessFile  *file,
 void
 gdm_display_access_file_close (GdmDisplayAccessFile  *file)
 {
+        char *auth_dir;
+
         g_return_if_fail (file != NULL);
         g_return_if_fail (file->priv->fp != NULL);
         g_return_if_fail (file->priv->path != NULL);
 
-        g_unlink (file->priv->path);
-        if (file->priv->path != NULL) {
-                g_free (file->priv->path);
-                file->priv->path = NULL;
-                g_object_notify (G_OBJECT (file), "path");
+        errno = 0;
+        if (g_unlink (file->priv->path) != 0) {
+                g_warning ("GdmDisplayAccessFile: Unable to remove X11 authority database '%s': %s",
+                           file->priv->path,
+                           g_strerror (errno));
         }
 
+        /* still try to remove dir even if file remove failed,
+           may have already been removed by someone else */
+        /* we own the parent directory too */
+        auth_dir = g_path_get_dirname (file->priv->path);
+        if (auth_dir != NULL) {
+                errno = 0;
+                if (g_rmdir (auth_dir) != 0) {
+                        g_warning ("GdmDisplayAccessFile: Unable to remove X11 authority directory '%s': %s",
+                                   auth_dir,
+                                   g_strerror (errno));
+                }
+                g_free (auth_dir);
+        }
+
+        g_free (file->priv->path);
+        file->priv->path = NULL;
+        g_object_notify (G_OBJECT (file), "path");
+
         fclose (file->priv->fp);
         file->priv->fp = NULL;
 }
